from typing import Optional


class AverageMeter:
    """
    Compute and stores the average and current value
    """

    val: float
    avg: float
    sum: float
    count: float

    def __init__(self, batch_size: Optional[float] = 1):
        self.batch_size = batch_size
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, batch_size: Optional[float] = None):
        batch_size = batch_size if batch_size is not None else self.batch_size
        self.val = val
        self.sum += val * batch_size
        self.count += batch_size
        self.avg = self.sum / self.count

    def __repr__(self):
        return repr(self.avg)



from .transformer import (FullyConnectedNetwork, MultiHeadAttention,
                          SelfAttention, TransformerEncoder,
                          TransformerEncoderLayer, UnitedPositionEmbedding)

__all__ = ['TransformerEncoder', 'TransformerEncoderLayer', 'MultiHeadAttention', 'SelfAttention', 'UnitedPositionEmbedding', 'FullyConnectedNetwork']



from __future__ import annotations

import atexit
import json
import logging
import logging.config
import os
import random
import shutil
from collections import OrderedDict
from collections.abc import MutableMapping
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import accelerate
import numpy as np
import torch
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torch.nn as nn
import torch.optim as optim
import torch.utils.data as data
from chanfig import Config

from danling.utils import catch, is_serializable

from .utils import ensure_dir, on_local_main_process, on_main_process


class BaseRunner(Config):
    """
    Set up everything for running a job
    """

    id: str
    name: str = 'danling'
    seed: int = 1031

    experiment_dir: str = 'experiments'
    checkpoint_dir_name: str = 'checkpoints'

    steps: int = 0
    epochs: int = 0

    accelerator: accelerate.Accelerator

    model: nn.Module
    optimizer: optim.Optimizer
    scheduler: optim.lr_scheduler._LRScheduler

    datasets: OrderedDict[str, data.DataLoader] = OrderedDict()
    datasamplers: OrderedDict[str, data.DataLoader] = OrderedDict()
    dataloaders: OrderedDict[str, data.DataLoader] = OrderedDict()
    batch_size: int = 1

    criterions: Tuple[nn.Module]

    results: List[Dict[str, Any]] = []
    result_best: Dict[str, Any] = {}
    result_last: Dict[str, Any] = {}
    score_best: float = 0
    score_last: float = 0

    log: bool = True
    logger: logging.Logger
    tensorboard: bool = False
    writer: Callable

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        # self.init_distributed()
        self.accelerator = accelerate.Accelerator(kwargs_handlers=[accelerate.DistributedDataParallelKwargs(find_unused_parameters=True)])

        self.init_seed()

        if getattr(self, 'deterministic', False):
            self.init_deterministic()

        if not hasattr(self, 'id'):
            self.id = f'{self.name}-{self.seed}'

        if getattr(self, 'log', False):
            self.init_logger()
            self.init_print()

        if getattr(self, 'tensorboard', False):
            self.writer = self.init_tensorboard()

        if self.is_main_process:
            self.yaml(os.path.join(self.dir, 'runner.yaml'))

        print(self.dict())
        atexit.register(self.print_result)

    def init_distributed(self) -> None:
        """
        Set up distributed training
        """
        dist.init_process_group(backend='nccl')
        self.process_index = dist.get_rank()
        self.num_processes = dist.get_world_size()
        self.local_process_index = int(os.environ.get("LOCAL_RANK", -1))
        torch.cuda.set_device(self.local_process_index)
        self.is_main_process = self.process_index == 0
        self.is_local_main_process = self.local_process_index == 0

    def init_deterministic(self) -> None:
        """
        Set up deterministic
        """
        cudnn.benchmark = False
        cudnn.deterministic = True

    def init_seed(self) -> None:
        """
        Set up random seed
        """
        if self.seed is None:
            self.seed = random.randint(0, 100000)
            if self.distributed:
                self.seed = self.accelerator.gather(torch.tensor(self.seed).cuda()).unsqueeze(0).flatten()[0]
        torch.manual_seed(self.seed + self.process_index)
        torch.cuda.manual_seed(self.seed + self.process_index)
        np.random.seed(self.seed + self.process_index)
        random.seed(self.seed + self.process_index)

    @on_main_process
    def init_logger(self) -> None:
        """
        Set up logger
        """
        # TODO consider move stream output out of stderr
        # Why is setting up proper logging so !@?#! ugly?
        logging.config.dictConfig({
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'standard': {
                    'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
                },
            },
            'handlers': {
                'stderr': {
                    'level': 'INFO',
                    'formatter': 'standard',
                    'class': 'logging.StreamHandler',
                    'stream': 'ext://sys.stderr',
                },
                'logfile': {
                    'level': 'DEBUG',
                    'formatter': 'standard',
                    'class': 'logging.FileHandler',
                    'filename': os.path.join(self.dir, 'run.log'),
                    'mode': 'a',
                }
            },
            'loggers': {
                '': {
                    'handlers': ['stderr', 'logfile'],
                    'level': 'DEBUG',
                    'propagate': True
                },
            }
        })
        logging.captureWarnings(True)
        self.logger = logging.getLogger('runner')
        self.logger.flush = lambda: [h.flush() for h in self.logger.handlers]

    def init_print(self, process: Optional[int] = 0, precision: Optional[int] = 10) -> None:
        """
        Set up print
        Only print from a specific process or force indicated
        Replace default print function with logging.info
        """
        torch.set_printoptions(precision=precision)

        logger = logging.getLogger('print')
        logger.flush = lambda: [h.flush for h in logger.handlers]
        import builtins as __builtin__
        builtin_print = __builtin__.print

        @catch()
        def print(*args, force=False, end='\n', file=None, flush=False, **kwargs):
            if self.process_index == process or force:
                if self.log:
                    logger.info(*args, **kwargs)
                else:
                    builtin_print(*args, end=end, file=file, flush=flush, **kwargs)

        __builtin__.print = print

    @on_main_process
    def init_tensorboard(self, *args, **kwargs):
        from torch.utils.tensorboard.writer import SummaryWriter
        return SummaryWriter(*args, **kwargs)

    def init_lr(self, lr_scale_factor: Optional[float] = None, batch_size_base: Optional[int] = None) -> None:
        """
        Set up learning rate
        """
        if lr_scale_factor is None:
            if batch_size_base is None:
                batch_size_base = self.batch_size_base
            lr_scale_factor = self.batch_size_actual / batch_size_base
        self.lr_scale_factor = lr_scale_factor
        self.lr = self.lr * self.lr_scale_factor
        self.lr_final = self.lr_final * self.lr_scale_factor

    @catch()
    @on_main_process
    def save(self) -> None:
        """
        Save checkpoint to checkpoint_dir
        """
        last_path = os.path.join(self.checkpoint_dir, 'last.pth')
        self.accelerator.save(self.dict(), last_path)
        if (self.epochs + 1) % self.save_freq == 0:
            save_path = os.path.join(self.checkpoint_dir, f'epoch-{self.epochs}.pth')
            shutil.copy(last_path, save_path)
        if self.is_best:
            best_path = os.path.join(self.checkpoint_dir, 'best.pth')
            shutil.copy(last_path, best_path)

    @catch()
    @on_main_process
    def save_result(self) -> None:
        """
        Save result
        """
        ret = {'id': self.id, 'name': self.name}
        ret.update(self.result_last)  # This is slower but ensure id in the first
        last_path = os.path.join(self.dir, 'last.json')
        with open(last_path, 'w') as f:
            json.dump(ret, f, indent=4)
        if self.is_best:
            best_path = os.path.join(self.dir, 'best.json')
            shutil.copy(last_path, best_path)

    def load(self, checkpoint: Union[MutableMapping, str]) -> None:
        """
        Load runner from checkpoint
        """
        print(f'=> loading checkpoint "{checkpoint}"')
        if isinstance(checkpoint, str):
            if not os.path.isfile(checkpoint):
                raise FileNotFoundError(f"checkpoint at {checkpoint} is not a file")
            checkpoint = torch.load(checkpoint)
        super().__init__(**checkpoint['runner'])
        if 'model' in checkpoint:
            self.model.load_state_dict(checkpoint['model'])
        if self.optimizer is not None and 'optimizer' in checkpoint:
            self.optimizer.load_state_dict(checkpoint['optimizer'])
        if self.scheduler is not None and 'scheduler' in checkpoint:
            self.scheduler.load_state_dict(checkpoint['scheduler'])
        print(f'=> loaded  checkpoint "{checkpoint}"')

    def dict(self, cls: Callable = dict) -> MutableMapping:
        dict = cls()
        for k, v in self.__dict__.items():
            if isinstance(v, Config):
                dict[k] = v.dict(cls)
            elif is_serializable(v):
                dict[k] = v
        return dict

    def state_dict(self, cls: Callable = OrderedDict) -> MutableMapping:
        """
        Return dict of all attributes for checkpoint
        """
        model = self.model
        if self.distributed:
            self.accelerator.wait_for_everyone()
            model = self.accelerator.unwrap_model(self.model)
        return cls(
            runner=self.dict(),
            model=model.state_dict(),
            optimizer=self.optimizer.state_dict() if self.optimizer else None,
            scheduler=self.scheduler.state_dict() if self.scheduler else None,
        )

    def step(self) -> None:
        """
        step optimizer and scheduler
        """
        if self.gradient_clip:
            nn.utils.clip_grad_norm_(self.model.parameters(), self.gradient_clip)
        if self.optimizer is not None:
            self.optimizer.step()
            self.optimizer.zero_grad()
        if self.scheduler is not None:
            self.scheduler.step()
        self.steps += 1

    def print_result(self) -> None:
        """
        Print last and best result
        """
        print(f'last result: {self.result_last}')
        print(f'best result: {self.result_best}')

    def __getattr__(self, name) -> Any:
        if (attr := getattr(self.accelerator, name, None)) is not None:
            return attr
        raise AttributeError(f"'Runner' object has no attribute '{name}'")

    def __repr__(self) -> str:
        return self.id

    def __str__(self) -> str:
        return self.name

    @property
    def distributed(self):
        """
        If runner is in distributed mode
        """
        return self.num_processes > 1

    @property
    def iters(self) -> int:
        """
        Number of iterations
        """
        return self.steps * self.batch_size_actual

    @property
    def progress(self) -> float:
        """
        Training Progress
        """
        if hasattr(self, 'iter_end'):
            return self.iters / self.iter_end
        elif hasattr(self, 'step_end'):
            return self.steps / self.step_end
        elif hasattr(self, 'epoch_end'):
            return self.epochs / self.epoch_end
        return 0

    @property
    def batch_size_actual(self) -> int:
        """
        Actual batch size
        """
        return self.batch_size * self.num_processes * getattr(self, 'accum_steps', 1)

    @property
    def is_best(self) -> bool:
        """
        If current result is best
        """
        return self.score_last > self.score_best

    @property
    @ensure_dir
    def dir(self) -> str:
        return os.path.join(self.experiment_dir, self.id)

    @property
    @ensure_dir
    def checkpoint_dir(self) -> str:
        return os.path.join(self.dir, self.checkpoint_dir_name)



from .encoder import TransformerEncoder, TransformerEncoderLayer
from .attention import MultiHeadAttention, SelfAttention
from .ffn import FullyConnectedNetwork
from .pos_embed import UnitedPositionEmbedding


__all__ = ['TransformerEncoder', 'TransformerEncoderLayer', 'MultiHeadAttention', 'SelfAttention', 'UnitedPositionEmbedding', 'FullyConnectedNetwork']



import torch


@torch.no_grad()
def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    ret = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
        ret.append(correct_k.mul_(100.0 / batch_size))
    return ret



from typing import Any, Dict, List, Optional, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


class FullyConnectedNetwork(nn.Module):

    def __init__(self, embed_dim: int, ffn_dim: int,
                 activation: Optional[str] = 'GELU',
                 feedforward_dropout: Optional[float] = 0.,
                 **kwargs: Optional[Dict[str, Any]]) -> None:
        super(FullyConnectedNetwork, self).__init__()
        self.fc1 = nn.Linear(embed_dim, ffn_dim)
        self.activation = getattr(nn, activation)()
        self.dropout = nn.Dropout(feedforward_dropout)
        self.fc2 = nn.Linear(ffn_dim, embed_dim)

    def forward(self, x: Tensor) -> Tensor:
        out = self.fc1(x)
        out = self.activation(out)
        out = self.dropout(out)
        out = self.fc2(out)
        return out



from math import ceil

import numpy as np
from torch.optim.lr_scheduler import _LRScheduler


class Scheduler(_LRScheduler):
    """
    Schedule the learning rate according to the learning rate schedule policy
    Support `cosine` and `linear` only
    Note that the scheduler schedule learning rate by steps
    """
    def __init__(
        self,
        optimizer,
        steps,
        lr_final=1e-6,
        lr_min=1e-6,
        policy='cosine',
        warmup_steps=10_000,
        accum_steps=1,
        last_epoch=-1,
    ):
        if policy not in ('cosine', 'linear'):
            raise ValueError(
                f'Only "cosine" or "linear" schedule policy are supported, but got {policy}')
        self.steps = ceil(steps / accum_steps)
        self.lr_final = lr_final
        self.lr_min = lr_min
        self.policy = policy
        self.method = getattr(self, self.policy)
        self.warmup_steps = ceil(warmup_steps / accum_steps)
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        progress = (self._step_count - self.warmup_steps) / float(self.steps - self.warmup_steps)
        progress = np.clip(progress, 0.0, 1.0)
        ratio = self.method(progress)
        if self.warmup_steps:
            ratio = ratio * np.minimum(1., self._step_count / self.warmup_steps)
        return [max(self.lr_min, lr * ratio) for lr in self.base_lrs]

    def linear(self, progress):
        return self.lr_final + (1 - self.lr_final) * (1.0 - progress)

    def cosine(self, progress):
        return 0.5 * (1. + np.cos(np.pi * progress))

    def __repr__(self):
        return f'{self.policy.capitalize()}Scheduler'



import json
import os
import pickle
import traceback
from functools import wraps
from typing import Any, Dict, List

import torch


def catch(error=Exception, exclude=None):
    """
    Catch error except for exclude
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except error as e:
                if exclude is not None and isinstance(e, exclude):
                    raise e
                trace = traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)
                print(''.join(trace), force=True)
                print(f'encoutered when calling {func} with args {args} and kwargs {kwargs}', force=True)
        return wrapper
    return decorator


def load(path: str, *args: List[Any], **kwargs: Dict[str, Any]) -> Any:
    """
    Load everything
    """
    if not os.path.isfile(path):
        raise ValueError(f'Trying to load {path} but it is not a file.')
    extension = os.path.splitext(path)[-1].lower()
    if extension == '.json':
        with open(path, 'r') as f:
            result = json.load(f, *args, **kwargs)
    elif extension == '.csv':
        import pandas as pd
        result = pd.read_csv(path, *args, **kwargs)
    elif extension == '.pth':
        result = torch.load(path)
    elif extension == '.npy' or extension == '.npz':
        import numpy as np
        result = np.load(path, allow_pickle=True)
    elif extension == '.pkl':
        try:
            try:
                with open(path, 'r') as f:
                    result = pickle.load(f, *args, **kwargs)
            except UnicodeDecodeError:
                with open(path, 'rb') as f:
                    result = pickle.load(f, *args, **kwargs)
        except:
            import pandas as pd
            result = pd.read_pickle(path, *args, **kwargs)
    else:
        raise ValueError(f'Tying to load {path} with unsupported extension')
    return result


def is_serializable(obj: Any) -> bool:
    """
    Check if obj is serializable
    """
    try:
        json.dumps(obj)
        return True
    except (TypeError, OverflowError):
        return False



from functools import partial
from typing import Any, Dict, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.utils.checkpoint import checkpoint

from .attention import MultiHeadAttention
from .ffn import FullyConnectedNetwork


class TransformerEncoderLayer(nn.Module):
    r"""TransformerEncoderLayer is made up of self-attn and feedforward network.
    This standard encoder layer is based on the paper "Attention Is All You Need".
    Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez,
    Lukasz Kaiser, and Illia Polosukhin. 2017. Attention is all you need. In Advances in
    Neural Information Processing Systems, pages 6000-6010. Users may modify or implement
    in a different way during application.
    Args:
        embed_dim: the number of expected features in the input (required).
        num_heads: the number of heads in the multi head attention models (required).
        ffn_dim: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of intermediate layer, relu or gelu (default=relu).
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False``.
        norm_first: if ``True``, layer norm is done prior to attention and feedforward
            operations, respectivaly. Otherwise it's done after. Default: ``False`` (after).
    Examples::
        >>> encoder_layer = nn.TransformerEncoderLayer(embed_dim=512, num_heads=8)
        >>> src = torch.rand(10, 32, 512)
        >>> out = encoder_layer(src)
    Alternatively, when ``batch_first`` is ``True``:
        >>> encoder_layer = nn.TransformerEncoderLayer(embed_dim=512, num_heads=8, batch_first=True)
        >>> src = torch.rand(32, 10, 512)
        >>> out = encoder_layer(src)
    """
    __constants__ = ['batch_first', 'norm_first']

    def __init__(self, embed_dim: int, num_heads: int,
                 ffn_dim: Optional[int] = 2048,
                 dropout: Optional[float] = 0.1,
                 attention_dropout: Optional[float] = 0.1,
                 feedforward_dropout: Optional[float] = 0.1,
                 activation: Optional[str] = 'GELU',
                 layer_norm_eps: Optional[float] = 1e-5,
                 scale_factor: Optional[float] = 1.,
                 bias: Optional[bool] = True,
                 add_bias_kv: Optional[bool] = False,
                 add_zero_attn: Optional[bool] = False,
                 batch_first: Optional[bool] = True,
                 norm_first: Optional[bool] = False,
                 Attention: Optional[nn.Module] = MultiHeadAttention,
                 FeedForwardNetwork: Optional[nn.Module] = FullyConnectedNetwork,
                 **kwargs: Optional[Dict[str, Any]]) -> None:
        super(TransformerEncoderLayer, self).__init__()
        self.norm_first = norm_first
        self.attention = Attention(
            embed_dim, num_heads, attention_dropout=attention_dropout,
            scale_factor=scale_factor, bias=bias, add_bias_kv=add_bias_kv,
            add_zero_attn=add_zero_attn, batch_first=batch_first, **kwargs)
        self.norm1 = nn.LayerNorm(embed_dim, eps=layer_norm_eps)
        self.feedforward = FeedForwardNetwork(
            embed_dim, ffn_dim, activation, feedforward_dropout, **kwargs)
        self.norm2 = nn.LayerNorm(embed_dim, eps=layer_norm_eps)
        self.dropout = nn.Dropout(dropout)

    def forward(self, src: Tensor,
                attn_bias: Optional[Tensor] = None,
                attn_mask: Optional[Tensor] = None,
                key_padding_mask: Optional[Tensor] = None,
                need_weights: Optional[bool] = False) -> Tensor:
        r"""Pass the input through the encoder layer.
        Args:
            src: the sequence to the encoder layer (required).
            attn_mask: the mask for the src sequence (optional).
            key_padding_mask: the mask for the src keys per batch (optional).
        Shape:
            see the docs in Transformer class.
        """

        if self.norm_first:
            src = self.norm1(src)

        attn, weights = self.attention(
            src, src, src, attn_bias=attn_bias, attn_mask=attn_mask,
            key_padding_mask=key_padding_mask, need_weights=need_weights)
        attn = src + self.dropout(attn)
        attn = self.norm1(attn) if not self.norm_first else self.norm2(attn)

        ffn = self.feedforward(attn)
        ffn = attn + self.dropout(ffn)

        if not self.norm_first:
            ffn = self.norm2(ffn)

        return ffn, weights


class TransformerEncoder(nn.Module):
    r"""TransformerEncoder is a stack of N encoder layers
    Args:
        num_layers: the number of sub-encoder-layers in the encoder (required).
        layer: the sub-encoder-layer in the encoder (default=TransformerEncoderLayer).
        drop_layer: the drop layer rate (default=0.0).
    Examples::
        >>> transformer_encoder = dl.model.TransformerEncoder(num_layers=6)
        >>> src = torch.rand(10, 32, 512)
        >>> out = transformer_encoder(src)
    """
    __constants__ = ['norm']

    def __init__(self,
                 layer: TransformerEncoderLayer,
                 num_layers: Optional[int] = 6,
                 **kwargs: Optional[Dict[str, Any]]) -> None:
        super(TransformerEncoder, self).__init__()
        self.num_layers = num_layers
        self.layers = nn.ModuleList([])
        self.layers.extend([layer for _ in range(self.num_layers)])

    def forward(self, src: Tensor, attn_bias: Optional[Tensor] = None,
                attn_mask: Optional[Tensor] = None,
                key_padding_mask: Optional[Tensor] = None,
                need_weights: Optional[bool] = False,
                gradient_checkpoint: Optional[bool] = False) -> Tensor:
        r"""Pass the input through the encoder layers in turn.
        Args:
            src: the sequence to the encoder (required).
            attn_mask: the mask for the src sequence (optional).
            key_padding_mask: the mask for the src keys per batch (optional).
        Shape:
            see the docs in Transformer class.
        """
        output = src
        attn_weights = [] if need_weights else torch.zeros(0, requires_grad=False)

        for layer in self.layers:
            if gradient_checkpoint and self.training:
                layer = partial(checkpoint, layer)
                need_weights = torch.tensor(need_weights)
            output, weights = layer(output, attn_bias, attn_mask, key_padding_mask, need_weights)
            if need_weights:
                attn_weights.append(weights)

        if need_weights:
            attn_weights = torch.stack(attn_weights).cpu().detach()

        return output, attn_weights



from functools import wraps
from os import makedirs
from os.path import abspath


def on_main_process(func):
    """
    Run func on main process
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.is_main_process or not self.distributed:
            return func(self, *args, **kwargs)
    return wrapper


def on_local_main_process(func):
    """
    Run func on local main process
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.is_local_main_process or not self.distributed:
            return func(self, *args, **kwargs)
    return wrapper


def ensure_dir(func):
    """
    Ensure directory exists
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        path = abspath(func(*args, **kwargs))
        makedirs(path, exist_ok=True)
        return path
    return wrapper



r"""
x: x center
y: y center
l: x left
r: x right
t: y top
b: y buttom
w: x width
h: y height
"""
import torch
from torch import Tensor


def ltrb2ltwh(bbox: Tensor) -> Tensor:
    l, t, r, b = bbox.T if bbox.ndim == 2 else bbox
    w, h = r - l, b - t
    return torch.stack((l, t, w, h)).T


def ltwh2ltrb(bbox: Tensor) -> Tensor:
    l, t, w, h = bbox.T if bbox.ndim == 2 else bbox
    r, b = l + w, t + h
    return torch.stack((l, t, r, b)).T


def ltrb2xywh(bbox: Tensor) -> Tensor:
    l, t, r, b = bbox.T if bbox.ndim == 2 else bbox
    w, h = r - l, b - t
    x, y = l + w / 2, t + h / 2
    return torch.stack((x, y, w, h)).T


def xywh2ltrb(bbox: Tensor) -> Tensor:
    x, y, w, h = bbox.T if bbox.ndim == 2 else bbox
    l, t, r, b = x - w / 2, y - h / 2, x + w / 2, y + h / 2
    return torch.stack((l, t, r, b)).T


def xywh2ltwh(bbox: Tensor) -> Tensor:
    x, y, w, h = bbox.T if bbox.ndim == 2 else bbox
    l, t = x - w / 2, y - h / 2
    return torch.stack((l, t, w, h)).T



from setuptools import find_packages, setup

setup(
    name='danling',
    version='0.0.1a3',
    description='danling',
    author='Zhiyuan Chen',
    author_email='this@zyc.ai',
    url='http://github.com/ZhiyuanChen/danling',
    packages=find_packages(),
)



from .metrics import AverageMeter
from .models import (FullyConnectedNetwork, MultiHeadAttention, SelfAttention,
                     TransformerEncoder, TransformerEncoderLayer,
                     UnitedPositionEmbedding)
from .runner import BaseRunner
from .utils import Scheduler, catch, load

__all__ = ['AverageMeter', 'Scheduler', 'TransformerEncoder', 'TransformerEncoderLayer', 'MultiHeadAttention', 'SelfAttention', 'UnitedPositionEmbedding', 'FullyConnectedNetwork', 'BaseRunner', 'catch', 'load']
